#pragma once

#include "drake/common/eigen_types.h"

// clang-format off
// NOLINTNEXTLINE(whitespace/line_length)
#warning "DRAKE DEPRECATED: The drake/common/ssize.h header is deprecated for removal. Use std::ssize instead. The deprecated code will be removed from Drake on or after 2026-09-01."
// clang-format on
