import collections.abc
import nanobind
import pydrake.common
import pydrake.math
import pydrake.symbolic
from _typeshed import Incomplete
from pydrake.common.cpp_template import DeterministicVector as DeterministicVector, GaussianVector as GaussianVector, UniformVector as UniformVector
from typing import ClassVar, overload

GetDeterministicValue: nanobind.nb_func
IsDeterministic: nanobind.nb_func
Mean: nanobind.nb_func
Sample: nanobind.nb_func
ToDistribution: nanobind.nb_func
ToDistributionVector: nanobind.nb_func
ToSymbolic: nanobind.nb_func

class Deterministic(Distribution):
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: Deterministic) -> None: ...
    @overload
    def __init__(self, value: float) -> None: ...
    def __copy__(self) -> Deterministic: ...
    def __deepcopy__(self, arg: dict) -> Deterministic: ...
    @property
    def __fields__(self): ...

class DeterministicVectorX(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorTNoneU) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorTNoneU: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣1𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT1U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT1U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣2𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT2U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT2U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣3𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT3U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT3U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣4𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT4U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT4U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣5𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT5U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT5U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣6𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT6U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT6U: ...
    @property
    def __fields__(self): ...

class DeterministicVector𝓣None𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorTNoneU) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorTNoneU: ...
    @property
    def __fields__(self): ...

class Distribution:
    def __init__(self, *args, **kwargs) -> None: ...
    def Mean(self) -> float: ...
    def Sample(self, generator: pydrake.common.RandomGenerator) -> float: ...
    def ToSymbolic(self) -> pydrake.symbolic.Expression: ...

class DistributionVector:
    def __init__(self, *args, **kwargs) -> None: ...
    def Mean(self, *args, **kwargs): ...
    def Sample(self, *args, **kwargs): ...
    def ToSymbolic(self, *args, **kwargs): ...

class Gaussian(Distribution):
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: Gaussian) -> None: ...
    @overload
    def __init__(self, mean: float, stddev: float) -> None: ...
    def __copy__(self) -> Gaussian: ...
    def __deepcopy__(self, arg: dict) -> Gaussian: ...
    @property
    def __fields__(self): ...

class GaussianVectorX(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorTNoneU) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorTNoneU: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣1𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT1U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT1U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣2𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT2U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT2U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣3𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT3U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT3U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣4𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT4U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT4U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣5𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT5U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT5U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣6𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT6U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT6U: ...
    @property
    def __fields__(self): ...

class GaussianVector𝓣None𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorTNoneU) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorTNoneU: ...
    @property
    def __fields__(self): ...

class Rotation:
    class AngleAxis:
        angle_deg: Incomplete
        axis: Incomplete
        @overload
        def __init__(self, other: Rotation.AngleAxis) -> None: ...
        @overload
        def __init__(self, **kwargs) -> None: ...
        def __copy__(self) -> Rotation.AngleAxis: ...
        def __deepcopy__(self, arg: dict) -> Rotation.AngleAxis: ...
        @property
        def __fields__(self): ...

    class Identity:
        @overload
        def __init__(self, other: Rotation.Identity) -> None: ...
        @overload
        def __init__(self, **kwargs) -> None: ...
        def __copy__(self) -> Rotation.Identity: ...
        def __deepcopy__(self, arg: dict) -> Rotation.Identity: ...
        @property
        def __fields__(self): ...

    class Rpy:
        deg: Incomplete
        @overload
        def __init__(self, other: Rotation.Rpy) -> None: ...
        @overload
        def __init__(self, **kwargs) -> None: ...
        def __copy__(self) -> Rotation.Rpy: ...
        def __deepcopy__(self, arg: dict) -> Rotation.Rpy: ...
        @property
        def __fields__(self): ...

    class Uniform:
        @overload
        def __init__(self, other: Rotation.Uniform) -> None: ...
        @overload
        def __init__(self, **kwargs) -> None: ...
        def __copy__(self) -> Rotation.Uniform: ...
        def __deepcopy__(self, arg: dict) -> Rotation.Uniform: ...
        @property
        def __fields__(self): ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: Rotation) -> None: ...
    @overload
    def __init__(self, arg: pydrake.math.RotationMatrix) -> None: ...
    @overload
    def __init__(self, **kwargs) -> None: ...
    def GetDeterministicValue(self) -> pydrake.math.RotationMatrix: ...
    def IsDeterministic(self) -> bool: ...
    def ToSymbolic(self) -> pydrake.math.RotationMatrix_TExpressionU: ...
    def set_rpy_deg(self, rpy_deg) -> None: ...
    def __copy__(self) -> Rotation: ...
    def __deepcopy__(self, arg: dict) -> Rotation: ...
    def __getattr__(self, arg: str) -> object: ...
    def __setattr__(self, arg0: str, arg1: object) -> None: ...
    @property
    def __fields__(self): ...

class Transform:
    _rewrite_yaml_dump_attr_name: ClassVar[nanobind.nb_func] = ...
    base_frame: Incomplete
    rotation: Rotation
    translation: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: Transform) -> None: ...
    @overload
    def __init__(self, arg: pydrake.math.RigidTransform) -> None: ...
    @overload
    def __init__(self, **kwargs) -> None: ...
    def GetDeterministicValue(self) -> pydrake.math.RigidTransform: ...
    def IsDeterministic(self) -> bool: ...
    def Mean(self) -> pydrake.math.RigidTransform: ...
    def Sample(self, generator: pydrake.common.RandomGenerator) -> pydrake.math.RigidTransform: ...
    def ToSymbolic(self) -> pydrake.math.RigidTransform_TExpressionU: ...
    def set_rotation_rpy_deg(self, rpy_deg) -> None: ...
    def __copy__(self) -> Transform: ...
    def __deepcopy__(self, arg: dict) -> Transform: ...
    @property
    def __fields__(self): ...

class Uniform(Distribution):
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: Uniform) -> None: ...
    @overload
    def __init__(self, min: float, max: float) -> None: ...
    def __copy__(self) -> Uniform: ...
    def __deepcopy__(self, arg: dict) -> Uniform: ...
    @property
    def __fields__(self): ...

class UniformDiscrete(Distribution):
    values: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformDiscrete) -> None: ...
    @overload
    def __init__(self, values: collections.abc.Sequence[float]) -> None: ...
    def __copy__(self) -> UniformDiscrete: ...
    def __deepcopy__(self, arg: dict) -> UniformDiscrete: ...
    @property
    def __fields__(self): ...

class UniformVectorX(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorTNoneU) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorTNoneU: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣1𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT1U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT1U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣2𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT2U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT2U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣3𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT3U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT3U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣4𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT4U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT4U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣5𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT5U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT5U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣6𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT6U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT6U: ...
    @property
    def __fields__(self): ...

class UniformVector𝓣None𝓤(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorTNoneU) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorTNoneU: ...
    @property
    def __fields__(self): ...

class _InvalidVariantSelectionDeterministic:
    def __init__(self, *args, **kwargs) -> None: ...

class _InvalidVariantSelectionGaussian:
    def __init__(self, *args, **kwargs) -> None: ...

class _InvalidVariantSelectionUniform:
    def __init__(self, *args, **kwargs) -> None: ...

class _TemporaryName_N5drake6schema13UniformVectorILi1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT1U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT1U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILi2EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT2U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT2U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILi3EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT3U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT3U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILi4EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT4U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT4U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILi5EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT5U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT5U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILi6EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorT6U) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorT6U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema13UniformVectorILin1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    max: Incomplete
    min: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: UniformVectorTNoneU) -> None: ...
    @overload
    def __init__(self, min, max) -> None: ...
    def __copy__(self) -> UniformVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> UniformVectorTNoneU: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT1U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT1U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi2EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT2U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT2U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi3EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT3U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT3U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi4EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT4U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT4U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi5EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT5U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT5U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILi6EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorT6U) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorT6U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema14GaussianVectorILin1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    mean: Incomplete
    stddev: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: GaussianVectorTNoneU) -> None: ...
    @overload
    def __init__(self, mean, stddev) -> None: ...
    def __copy__(self) -> GaussianVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> GaussianVectorTNoneU: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT1U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT1U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT1U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi2EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT2U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT2U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT2U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi3EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT3U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT3U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT3U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi4EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT4U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT4U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT4U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi5EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT5U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT5U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT5U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILi6EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorT6U) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorT6U: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorT6U: ...
    @property
    def __fields__(self): ...

class _TemporaryName_N5drake6schema19DeterministicVectorILin1EEE(DistributionVector):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    value: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: DeterministicVectorTNoneU) -> None: ...
    @overload
    def __init__(self, value) -> None: ...
    def __copy__(self) -> DeterministicVectorTNoneU: ...
    def __deepcopy__(self, arg: dict) -> DeterministicVectorTNoneU: ...
    @property
    def __fields__(self): ...
