import nanobind
import pydrake.autodiffutils
import pydrake.symbolic
from _typeshed import Incomplete
from pydrake.common.cpp_template import AngleAxis_ as AngleAxis_, Isometry3_ as Isometry3_, Quaternion_ as Quaternion_
from typing import Callable, ClassVar, overload

__getattr__: Callable

class AngleAxis:
    Identity: ClassVar[nanobind.nb_func] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast𝓣float𝓤: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, angle: float, axis) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: AngleAxis) -> None: ...
    def angle(self) -> float: ...
    def axis(self, *args, **kwargs): ...
    def inverse(self) -> AngleAxis: ...
    def multiply(self, other: AngleAxis) -> Quaternion: ...
    def quaternion(self) -> Quaternion: ...
    def rotation(self, *args, **kwargs): ...
    def set_angle(self, angle: float) -> None: ...
    def set_axis(self, axis) -> None: ...
    def set_quaternion(self, q: Quaternion) -> None: ...
    def set_rotation(self, rotation) -> None: ...
    def __copy__(self) -> AngleAxis: ...
    def __deepcopy__(self, arg: dict) -> AngleAxis: ...
    def __matmul__(self, *args, **kwargs): ...

class AngleAxis_𝓣AutoDiffXd𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, angle: pydrake.autodiffutils.AutoDiffXd, axis) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: AngleAxis_TAutoDiffXdU) -> None: ...
    def angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def axis(self, *args, **kwargs): ...
    def inverse(self) -> AngleAxis_TAutoDiffXdU: ...
    def multiply(self, other: AngleAxis_TAutoDiffXdU) -> Quaternion_TAutoDiffXdU: ...
    def quaternion(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_angle(self, angle: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_axis(self, axis) -> None: ...
    def set_quaternion(self, q: Quaternion_TAutoDiffXdU) -> None: ...
    def set_rotation(self, rotation) -> None: ...
    def __copy__(self) -> AngleAxis_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> AngleAxis_TAutoDiffXdU: ...
    def __matmul__(self, *args, **kwargs): ...

class AngleAxis_𝓣Expression𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, angle: pydrake.symbolic.Expression, axis) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TExpressionU) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: AngleAxis_TExpressionU) -> None: ...
    def angle(self) -> pydrake.symbolic.Expression: ...
    def axis(self, *args, **kwargs): ...
    def inverse(self) -> AngleAxis_TExpressionU: ...
    def multiply(self, other: AngleAxis_TExpressionU) -> Quaternion_TExpressionU: ...
    def quaternion(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_angle(self, angle: pydrake.symbolic.Expression) -> None: ...
    def set_axis(self, axis) -> None: ...
    def set_quaternion(self, q: Quaternion_TExpressionU) -> None: ...
    def set_rotation(self, rotation) -> None: ...
    def __copy__(self) -> AngleAxis_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> AngleAxis_TExpressionU: ...
    def __matmul__(self, *args, **kwargs): ...

class Isometry3:
    Identity: ClassVar[nanobind.nb_func] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast𝓣float𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    @overload
    def __init__(self, rotation, translation) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion, translation) -> None: ...
    @overload
    def __init__(self, other: Isometry3) -> None: ...
    def inverse(self) -> Isometry3: ...
    def matrix(self, *args, **kwargs): ...
    def quaternion(self) -> Quaternion: ...
    def rotation(self, *args, **kwargs): ...
    def set_matrix(self, arg) -> None: ...
    def set_quaternion(self, arg: Quaternion) -> None: ...
    def set_rotation(self, arg) -> None: ...
    def set_translation(self, arg) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> Isometry3: ...
    def __deepcopy__(self, arg: dict) -> Isometry3: ...

class Isometry3_𝓣AutoDiffXd𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    @overload
    def __init__(self, rotation, translation) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TAutoDiffXdU, translation) -> None: ...
    @overload
    def __init__(self, other: Isometry3_TAutoDiffXdU) -> None: ...
    def inverse(self) -> Isometry3_TAutoDiffXdU: ...
    def matrix(self, *args, **kwargs): ...
    def quaternion(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_matrix(self, arg) -> None: ...
    def set_quaternion(self, arg: Quaternion_TAutoDiffXdU) -> None: ...
    def set_rotation(self, arg) -> None: ...
    def set_translation(self, arg) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> Isometry3_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> Isometry3_TAutoDiffXdU: ...

class Isometry3_𝓣Expression𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    @overload
    def __init__(self, rotation, translation) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TExpressionU, translation) -> None: ...
    @overload
    def __init__(self, other: Isometry3_TExpressionU) -> None: ...
    def inverse(self) -> Isometry3_TExpressionU: ...
    def matrix(self, *args, **kwargs): ...
    def quaternion(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_matrix(self, arg) -> None: ...
    def set_quaternion(self, arg: Quaternion_TExpressionU) -> None: ...
    def set_rotation(self, arg) -> None: ...
    def set_translation(self, arg) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> Isometry3_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> Isometry3_TExpressionU: ...

class Quaternion:
    Identity: ClassVar[nanobind.nb_func] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast𝓣float𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, wxyz) -> None: ...
    @overload
    def __init__(self, w: float, x: float, y: float, z: float) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: Quaternion) -> None: ...
    def conjugate(self) -> Quaternion: ...
    def inverse(self) -> Quaternion: ...
    def rotation(self, *args, **kwargs): ...
    def set_rotation(self, arg) -> None: ...
    @overload
    def set_wxyz(self, wxyz) -> None: ...
    @overload
    def set_wxyz(self, w: float, x: float, y: float, z: float) -> None: ...
    def slerp(self, t: float, other: Quaternion) -> Quaternion: ...
    def w(self) -> float: ...
    def wxyz(self, *args, **kwargs): ...
    def x(self) -> float: ...
    def xyz(self, *args, **kwargs): ...
    def y(self) -> float: ...
    def z(self) -> float: ...
    def __copy__(self) -> Quaternion: ...
    def __deepcopy__(self, arg: dict) -> Quaternion: ...

class Quaternion_𝓣AutoDiffXd𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, wxyz) -> None: ...
    @overload
    def __init__(self, w: pydrake.autodiffutils.AutoDiffXd, x: pydrake.autodiffutils.AutoDiffXd, y: pydrake.autodiffutils.AutoDiffXd, z: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: Quaternion_TAutoDiffXdU) -> None: ...
    def conjugate(self) -> Quaternion_TAutoDiffXdU: ...
    def inverse(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_rotation(self, arg) -> None: ...
    @overload
    def set_wxyz(self, wxyz) -> None: ...
    @overload
    def set_wxyz(self, w: pydrake.autodiffutils.AutoDiffXd, x: pydrake.autodiffutils.AutoDiffXd, y: pydrake.autodiffutils.AutoDiffXd, z: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def slerp(self, t: float, other: Quaternion_TAutoDiffXdU) -> Quaternion_TAutoDiffXdU: ...
    def w(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def wxyz(self, *args, **kwargs): ...
    def x(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def xyz(self, *args, **kwargs): ...
    def y(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def z(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> Quaternion_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> Quaternion_TAutoDiffXdU: ...

class Quaternion_𝓣Expression𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, wxyz) -> None: ...
    @overload
    def __init__(self, w: pydrake.symbolic.Expression, x: pydrake.symbolic.Expression, y: pydrake.symbolic.Expression, z: pydrake.symbolic.Expression) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: Quaternion_TExpressionU) -> None: ...
    def conjugate(self) -> Quaternion_TExpressionU: ...
    def inverse(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_rotation(self, arg) -> None: ...
    @overload
    def set_wxyz(self, wxyz) -> None: ...
    @overload
    def set_wxyz(self, w: pydrake.symbolic.Expression, x: pydrake.symbolic.Expression, y: pydrake.symbolic.Expression, z: pydrake.symbolic.Expression) -> None: ...
    def slerp(self, t: float, other: Quaternion_TExpressionU) -> Quaternion_TExpressionU: ...
    def w(self) -> pydrake.symbolic.Expression: ...
    def wxyz(self, *args, **kwargs): ...
    def x(self) -> pydrake.symbolic.Expression: ...
    def xyz(self, *args, **kwargs): ...
    def y(self) -> pydrake.symbolic.Expression: ...
    def z(self) -> pydrake.symbolic.Expression: ...
    def __copy__(self) -> Quaternion_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> Quaternion_TExpressionU: ...

class _TemporaryName_N5Eigen10QuaternionIN5drake2ad8AutoDiffELi0EEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, wxyz) -> None: ...
    @overload
    def __init__(self, w: pydrake.autodiffutils.AutoDiffXd, x: pydrake.autodiffutils.AutoDiffXd, y: pydrake.autodiffutils.AutoDiffXd, z: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: Quaternion_TAutoDiffXdU) -> None: ...
    def conjugate(self) -> Quaternion_TAutoDiffXdU: ...
    def inverse(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_rotation(self, arg) -> None: ...
    @overload
    def set_wxyz(self, wxyz) -> None: ...
    @overload
    def set_wxyz(self, w: pydrake.autodiffutils.AutoDiffXd, x: pydrake.autodiffutils.AutoDiffXd, y: pydrake.autodiffutils.AutoDiffXd, z: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def slerp(self, t: float, other: Quaternion_TAutoDiffXdU) -> Quaternion_TAutoDiffXdU: ...
    def w(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def wxyz(self, *args, **kwargs): ...
    def x(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def xyz(self, *args, **kwargs): ...
    def y(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def z(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> Quaternion_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> Quaternion_TAutoDiffXdU: ...

class _TemporaryName_N5Eigen10QuaternionIN5drake8symbolic10ExpressionELi0EEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, wxyz) -> None: ...
    @overload
    def __init__(self, w: pydrake.symbolic.Expression, x: pydrake.symbolic.Expression, y: pydrake.symbolic.Expression, z: pydrake.symbolic.Expression) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: Quaternion_TExpressionU) -> None: ...
    def conjugate(self) -> Quaternion_TExpressionU: ...
    def inverse(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_rotation(self, arg) -> None: ...
    @overload
    def set_wxyz(self, wxyz) -> None: ...
    @overload
    def set_wxyz(self, w: pydrake.symbolic.Expression, x: pydrake.symbolic.Expression, y: pydrake.symbolic.Expression, z: pydrake.symbolic.Expression) -> None: ...
    def slerp(self, t: float, other: Quaternion_TExpressionU) -> Quaternion_TExpressionU: ...
    def w(self) -> pydrake.symbolic.Expression: ...
    def wxyz(self, *args, **kwargs): ...
    def x(self) -> pydrake.symbolic.Expression: ...
    def xyz(self, *args, **kwargs): ...
    def y(self) -> pydrake.symbolic.Expression: ...
    def z(self) -> pydrake.symbolic.Expression: ...
    def __copy__(self) -> Quaternion_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> Quaternion_TExpressionU: ...

class _TemporaryName_N5Eigen9AngleAxisIN5drake2ad8AutoDiffEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, angle: pydrake.autodiffutils.AutoDiffXd, axis) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: AngleAxis_TAutoDiffXdU) -> None: ...
    def angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def axis(self, *args, **kwargs): ...
    def inverse(self) -> AngleAxis_TAutoDiffXdU: ...
    def multiply(self, other: AngleAxis_TAutoDiffXdU) -> Quaternion_TAutoDiffXdU: ...
    def quaternion(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_angle(self, angle: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    def set_axis(self, axis) -> None: ...
    def set_quaternion(self, q: Quaternion_TAutoDiffXdU) -> None: ...
    def set_rotation(self, rotation) -> None: ...
    def __copy__(self) -> AngleAxis_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> AngleAxis_TAutoDiffXdU: ...
    def __matmul__(self, *args, **kwargs): ...

class _TemporaryName_N5Eigen9AngleAxisIN5drake8symbolic10ExpressionEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, angle: pydrake.symbolic.Expression, axis) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TExpressionU) -> None: ...
    @overload
    def __init__(self, rotation) -> None: ...
    @overload
    def __init__(self, other: AngleAxis_TExpressionU) -> None: ...
    def angle(self) -> pydrake.symbolic.Expression: ...
    def axis(self, *args, **kwargs): ...
    def inverse(self) -> AngleAxis_TExpressionU: ...
    def multiply(self, other: AngleAxis_TExpressionU) -> Quaternion_TExpressionU: ...
    def quaternion(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_angle(self, angle: pydrake.symbolic.Expression) -> None: ...
    def set_axis(self, axis) -> None: ...
    def set_quaternion(self, q: Quaternion_TExpressionU) -> None: ...
    def set_rotation(self, rotation) -> None: ...
    def __copy__(self) -> AngleAxis_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> AngleAxis_TExpressionU: ...
    def __matmul__(self, *args, **kwargs): ...

class _TemporaryName_N5Eigen9TransformIN5drake2ad8AutoDiffELi3ELi1ELi0EEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    @overload
    def __init__(self, rotation, translation) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TAutoDiffXdU, translation) -> None: ...
    @overload
    def __init__(self, other: Isometry3_TAutoDiffXdU) -> None: ...
    def inverse(self) -> Isometry3_TAutoDiffXdU: ...
    def matrix(self, *args, **kwargs): ...
    def quaternion(self) -> Quaternion_TAutoDiffXdU: ...
    def rotation(self, *args, **kwargs): ...
    def set_matrix(self, arg) -> None: ...
    def set_quaternion(self, arg: Quaternion_TAutoDiffXdU) -> None: ...
    def set_rotation(self, arg) -> None: ...
    def set_translation(self, arg) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> Isometry3_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> Isometry3_TAutoDiffXdU: ...

class _TemporaryName_N5Eigen9TransformIN5drake8symbolic10ExpressionELi3ELi1ELi0EEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    @overload
    def __init__(self, rotation, translation) -> None: ...
    @overload
    def __init__(self, quaternion: Quaternion_TExpressionU, translation) -> None: ...
    @overload
    def __init__(self, other: Isometry3_TExpressionU) -> None: ...
    def inverse(self) -> Isometry3_TExpressionU: ...
    def matrix(self, *args, **kwargs): ...
    def quaternion(self) -> Quaternion_TExpressionU: ...
    def rotation(self, *args, **kwargs): ...
    def set_matrix(self, arg) -> None: ...
    def set_quaternion(self, arg: Quaternion_TExpressionU) -> None: ...
    def set_rotation(self, arg) -> None: ...
    def set_translation(self, arg) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> Isometry3_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> Isometry3_TExpressionU: ...
