import collections.abc
import enum
import nanobind
import pydrake.autodiffutils
import pydrake.common.eigen_geometry
import pydrake.symbolic
import pydrake.symbolic as _sym
from _typeshed import Incomplete
from pydrake.common.cpp_template import BsplineBasis_ as BsplineBasis_, RigidTransform_ as RigidTransform_, RollPitchYaw_ as RollPitchYaw_, RotationMatrix_ as RotationMatrix_
from typing import Callable, ClassVar, overload

AreQuaternionsEqualForOrientation: nanobind.nb_func
BalanceQuadraticForms: nanobind.nb_func
CalculateAngularVelocityExpressedInBFromQuaternionDt: nanobind.nb_func
CalculateQuaternionDtConstraintViolation: nanobind.nb_func
CalculateQuaternionDtFromAngularVelocityExpressedInB: nanobind.nb_func
ClosestQuaternion: nanobind.nb_func
ComputeNumericalGradient: nanobind.nb_func
ContinuousAlgebraicRiccatiEquation: nanobind.nb_func
DecomposePSDmatrixIntoXtransposeTimesX: nanobind.nb_func
DecomposePositiveQuadraticForm: nanobind.nb_func
DiscreteAlgebraicRiccatiEquation: nanobind.nb_func
ExtractPrincipalSubmatrix: nanobind.nb_func
IsBothQuaternionAndQuaternionDtOK: nanobind.nb_func
IsPositiveDefinite: nanobind.nb_func
IsQuaternionValid: nanobind.nb_func
IsSymmetric: nanobind.nb_func
QuaternionToCanonicalForm: nanobind.nb_func
RealContinuousLyapunovEquation: nanobind.nb_func
RealDiscreteLyapunovEquation: nanobind.nb_func
SoftOverMax: nanobind.nb_func
SoftOverMin: nanobind.nb_func
SoftUnderMax: nanobind.nb_func
SoftUnderMin: nanobind.nb_func
ToLowerTriangularColumnsFromMatrix: nanobind.nb_func
ToSymmetricMatrixFromLowerTriangularColumns: nanobind.nb_func
UniformlyRandomAngleAxis: nanobind.nb_func
UniformlyRandomQuaternion: nanobind.nb_func
UniformlyRandomRPY: nanobind.nb_func
UniformlyRandomRotationMatrix: nanobind.nb_func
VectorToSkewSymmetric: nanobind.nb_func
__getattr__: Callable
abs: nanobind.nb_func
acos: nanobind.nb_func
arccos: nanobind.nb_func
arcsin: nanobind.nb_func
arctan: nanobind.nb_func
arctan2: nanobind.nb_func
asin: nanobind.nb_func
atan: nanobind.nb_func
atan2: nanobind.nb_func
ceil: nanobind.nb_func
cos: nanobind.nb_func
cosh: nanobind.nb_func
exp: nanobind.nb_func
floor: nanobind.nb_func
inv: nanobind.nb_func
is_quaternion_in_canonical_form: nanobind.nb_func
isnan: nanobind.nb_func
log: nanobind.nb_func
matmul: nanobind.nb_func
max: nanobind.nb_func
min: nanobind.nb_func
pow: nanobind.nb_func
sin: nanobind.nb_func
sinh: nanobind.nb_func
sqrt: nanobind.nb_func
tan: nanobind.nb_func
tanh: nanobind.nb_func
wrap_to: nanobind.nb_func

class BarycentricMesh:
    def __init__(self, arg: collections.abc.Sequence[collections.abc.Set[float]]) -> None: ...
    def Eval(self, *args, **kwargs): ...
    def EvalBarycentricWeights(self, *args, **kwargs): ...
    def MeshValuesFrom(self, *args, **kwargs): ...
    def get_all_mesh_points(self, *args, **kwargs): ...
    def get_input_grid(self) -> list[set[float]]: ...
    def get_input_size(self) -> int: ...
    def get_mesh_point(self, *args, **kwargs): ...
    def get_num_interpolants(self) -> int: ...
    def get_num_mesh_points(self) -> int: ...

class BsplineBasis:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, order: int, knots: collections.abc.Sequence[float]) -> None: ...
    @overload
    def __init__(self, order: int, num_basis_functions: int, type: KnotVectorType = ..., initial_parameter_value: float = ..., final_parameter_value: float = ...) -> None: ...
    @overload
    def __init__(self, other: BsplineBasis) -> None: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_interval: collections.abc.Sequence[float]) -> list[int]: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_value: float) -> list[int]: ...
    def EvaluateBasisFunctionI(self, i: int, parameter_value: float) -> float: ...
    def EvaluateCurve(self, *args, **kwargs): ...
    def EvaluateLinearInControlPoints(self, *args, **kwargs): ...
    def FindContainingInterval(self, parameter_value: float) -> int: ...
    def degree(self) -> int: ...
    def final_parameter_value(self) -> float: ...
    def initial_parameter_value(self) -> float: ...
    def knots(self) -> list[float]: ...
    def num_basis_functions(self) -> int: ...
    def order(self) -> int: ...
    def __eq__(self, arg: BsplineBasis) -> bool: ...
    def __ne__(self, arg: BsplineBasis) -> bool: ...

class BsplineBasis_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, order: int, knots: collections.abc.Sequence[pydrake.autodiffutils.AutoDiffXd]) -> None: ...
    @overload
    def __init__(self, order: int, num_basis_functions: int, type: KnotVectorType = ..., initial_parameter_value: pydrake.autodiffutils.AutoDiffXd = ..., final_parameter_value: pydrake.autodiffutils.AutoDiffXd = ...) -> None: ...
    @overload
    def __init__(self, other: BsplineBasis_TAutoDiffXdU) -> None: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_interval: collections.abc.Sequence[pydrake.autodiffutils.AutoDiffXd]) -> list[int]: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> list[int]: ...
    def EvaluateBasisFunctionI(self, i: int, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> pydrake.autodiffutils.AutoDiffXd: ...
    def EvaluateCurve(self, *args, **kwargs): ...
    def EvaluateLinearInControlPoints(self, *args, **kwargs): ...
    def FindContainingInterval(self, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> int: ...
    def degree(self) -> int: ...
    def final_parameter_value(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def initial_parameter_value(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def knots(self) -> list[pydrake.autodiffutils.AutoDiffXd]: ...
    def num_basis_functions(self) -> int: ...
    def order(self) -> int: ...
    def __eq__(self, arg: BsplineBasis_TAutoDiffXdU) -> bool: ...
    def __ne__(self, arg: BsplineBasis_TAutoDiffXdU) -> bool: ...

class BsplineBasis_𝓣Expression𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, order: int, knots: collections.abc.Sequence[pydrake.symbolic.Expression]) -> None: ...
    @overload
    def __init__(self, order: int, num_basis_functions: int, type: KnotVectorType = ..., initial_parameter_value: pydrake.symbolic.Expression = ..., final_parameter_value: pydrake.symbolic.Expression = ...) -> None: ...
    @overload
    def __init__(self, other: BsplineBasis_TExpressionU) -> None: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_interval: collections.abc.Sequence[pydrake.symbolic.Expression]) -> list[int]: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_value: pydrake.symbolic.Expression) -> list[int]: ...
    def EvaluateBasisFunctionI(self, i: int, parameter_value: pydrake.symbolic.Expression) -> pydrake.symbolic.Expression: ...
    def EvaluateCurve(self, *args, **kwargs): ...
    def EvaluateLinearInControlPoints(self, *args, **kwargs): ...
    def FindContainingInterval(self, parameter_value: pydrake.symbolic.Expression) -> int: ...
    def degree(self) -> int: ...
    def final_parameter_value(self) -> pydrake.symbolic.Expression: ...
    def initial_parameter_value(self) -> pydrake.symbolic.Expression: ...
    def knots(self) -> list[pydrake.symbolic.Expression]: ...
    def num_basis_functions(self) -> int: ...
    def order(self) -> int: ...
    def __eq__(self, arg: BsplineBasis_TExpressionU) -> pydrake.symbolic.Formula: ...
    def __ne__(self, arg: BsplineBasis_TExpressionU) -> pydrake.symbolic.Formula: ...

class KnotVectorType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _hashable_values_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _unhashable_values_map_: ClassVar[dict] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    kClampedUniform: ClassVar[KnotVectorType] = ...
    kUniform: ClassVar[KnotVectorType] = ...
    __nb_enum__: ClassVar[PyCapsule] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class NumericalGradientMethod(enum.Enum):
    __new__: ClassVar[Callable] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _hashable_values_: ClassVar[list] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[object]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _unhashable_values_map_: ClassVar[dict] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    _value_repr_: ClassVar[None] = ...
    kBackward: ClassVar[NumericalGradientMethod] = ...
    kCentral: ClassVar[NumericalGradientMethod] = ...
    kForward: ClassVar[NumericalGradientMethod] = ...
    __nb_enum__: ClassVar[PyCapsule] = ...

class NumericalGradientOption:
    def __init__(self, method: NumericalGradientMethod, function_accuracy: float = ...) -> None: ...
    def NumericalGradientMethod(self) -> NumericalGradientMethod: ...
    def perturbation_size(self) -> float: ...

class RigidTransform:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast𝓣float𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: RigidTransform) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix, p) -> None: ...
    @overload
    def __init__(self, rpy: RollPitchYaw, p) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion, p) -> None: ...
    @overload
    def __init__(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis, p) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix) -> None: ...
    @overload
    def __init__(self, p) -> None: ...
    @overload
    def __init__(self, pose: pydrake.common.eigen_geometry.Isometry3) -> None: ...
    @overload
    def __init__(self, pose) -> None: ...
    def GetAsIsometry3(self) -> pydrake.common.eigen_geometry.Isometry3: ...
    def GetAsMatrix34(self, *args, **kwargs): ...
    def GetAsMatrix4(self, *args, **kwargs): ...
    def GetMaximumAbsoluteDifference(self, other: RigidTransform) -> float: ...
    def GetMaximumAbsoluteTranslationDifference(self, other: RigidTransform) -> float: ...
    def InvertAndCompose(self, other: RigidTransform) -> RigidTransform: ...
    def IsExactlyEqualTo(self, other: RigidTransform) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RigidTransform, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, translation_tolerance: float) -> bool: ...
    def SetFromIsometry3(self, pose: pydrake.common.eigen_geometry.Isometry3) -> None: ...
    def SetIdentity(self) -> RigidTransform: ...
    def inverse(self) -> RigidTransform: ...
    def rotation(self) -> RotationMatrix: ...
    def set(self, R: RotationMatrix, p) -> None: ...
    @overload
    def set_rotation(self, R: RotationMatrix) -> None: ...
    @overload
    def set_rotation(self, rpy: RollPitchYaw) -> None: ...
    @overload
    def set_rotation(self, quaternion: pydrake.common.eigen_geometry.Quaternion) -> None: ...
    @overload
    def set_rotation(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis) -> None: ...
    def set_translation(self, p) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> RigidTransform: ...
    def __deepcopy__(self, arg: dict) -> RigidTransform: ...

class RigidTransform_𝓣AutoDiffXd𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: RigidTransform_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, rpy: RollPitchYaw_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, p) -> None: ...
    @overload
    def __init__(self, pose: pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, pose) -> None: ...
    def GetAsIsometry3(self) -> pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU: ...
    def GetAsMatrix34(self, *args, **kwargs): ...
    def GetAsMatrix4(self, *args, **kwargs): ...
    def GetMaximumAbsoluteDifference(self, other: RigidTransform_TAutoDiffXdU) -> pydrake.autodiffutils.AutoDiffXd: ...
    def GetMaximumAbsoluteTranslationDifference(self, other: RigidTransform_TAutoDiffXdU) -> pydrake.autodiffutils.AutoDiffXd: ...
    def InvertAndCompose(self, other: RigidTransform_TAutoDiffXdU) -> RigidTransform_TAutoDiffXdU: ...
    def IsExactlyEqualTo(self, other: RigidTransform_TAutoDiffXdU) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RigidTransform_TAutoDiffXdU, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, translation_tolerance: float) -> bool: ...
    def SetFromIsometry3(self, pose: pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU) -> None: ...
    def SetIdentity(self) -> RigidTransform_TAutoDiffXdU: ...
    def inverse(self) -> RigidTransform_TAutoDiffXdU: ...
    def rotation(self) -> RotationMatrix_TAutoDiffXdU: ...
    def set(self, R: RotationMatrix_TAutoDiffXdU, p) -> None: ...
    @overload
    def set_rotation(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, rpy: RollPitchYaw_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU) -> None: ...
    def set_translation(self, p) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> RigidTransform_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RigidTransform_TAutoDiffXdU: ...

class RigidTransform_𝓣Expression𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: RigidTransform_TExpressionU) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, rpy: RollPitchYaw_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def __init__(self, p) -> None: ...
    @overload
    def __init__(self, pose: pydrake.common.eigen_geometry.Isometry3_TExpressionU) -> None: ...
    @overload
    def __init__(self, pose) -> None: ...
    def GetAsIsometry3(self) -> pydrake.common.eigen_geometry.Isometry3_TExpressionU: ...
    def GetAsMatrix34(self, *args, **kwargs): ...
    def GetAsMatrix4(self, *args, **kwargs): ...
    def GetMaximumAbsoluteDifference(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Expression: ...
    def GetMaximumAbsoluteTranslationDifference(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Expression: ...
    def InvertAndCompose(self, other: RigidTransform_TExpressionU) -> RigidTransform_TExpressionU: ...
    def IsExactlyEqualTo(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Formula: ...
    def IsExactlyIdentity(self) -> pydrake.symbolic.Formula: ...
    def IsNearlyEqualTo(self, other: RigidTransform_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def IsNearlyIdentity(self, translation_tolerance: float) -> pydrake.symbolic.Formula: ...
    def SetFromIsometry3(self, pose: pydrake.common.eigen_geometry.Isometry3_TExpressionU) -> None: ...
    def SetIdentity(self) -> RigidTransform_TExpressionU: ...
    def inverse(self) -> RigidTransform_TExpressionU: ...
    def rotation(self) -> RotationMatrix_TExpressionU: ...
    def set(self, R: RotationMatrix_TExpressionU, p) -> None: ...
    @overload
    def set_rotation(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, rpy: RollPitchYaw_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TExpressionU) -> None: ...
    def set_translation(self, p) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> RigidTransform_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RigidTransform_TExpressionU: ...

class RollPitchYaw:
    @overload
    def __init__(self, other: RollPitchYaw) -> None: ...
    @overload
    def __init__(self, rpy) -> None: ...
    @overload
    def __init__(self, roll: float, pitch: float, yaw: float) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    def CalcAngularVelocityInChildFromRpyDt(self, *args, **kwargs): ...
    def CalcAngularVelocityInParentFromRpyDt(self, *args, **kwargs): ...
    def CalcRotationMatrixDt(self, *args, **kwargs): ...
    def CalcRpyDDtFromAngularAccelInChild(self, *args, **kwargs): ...
    def CalcRpyDDtFromRpyDtAndAngularAccelInParent(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInChild(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInParent(self, *args, **kwargs): ...
    def IsNearlyEqualTo(self, other: RollPitchYaw, tolerance: float) -> bool: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion: ...
    def ToRotationMatrix(self) -> RotationMatrix: ...
    def pitch_angle(self) -> float: ...
    def roll_angle(self) -> float: ...
    def vector(self, *args, **kwargs): ...
    def yaw_angle(self) -> float: ...
    def __copy__(self) -> RollPitchYaw: ...
    def __deepcopy__(self, arg: dict) -> RollPitchYaw: ...

class RollPitchYaw_𝓣AutoDiffXd𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self, other: RollPitchYaw_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, rpy) -> None: ...
    @overload
    def __init__(self, roll: pydrake.autodiffutils.AutoDiffXd, pitch: pydrake.autodiffutils.AutoDiffXd, yaw: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    def CalcAngularVelocityInChildFromRpyDt(self, *args, **kwargs): ...
    def CalcAngularVelocityInParentFromRpyDt(self, *args, **kwargs): ...
    def CalcRotationMatrixDt(self, *args, **kwargs): ...
    def CalcRpyDDtFromAngularAccelInChild(self, *args, **kwargs): ...
    def CalcRpyDDtFromRpyDtAndAngularAccelInParent(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInChild(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInParent(self, *args, **kwargs): ...
    def IsNearlyEqualTo(self, other: RollPitchYaw_TAutoDiffXdU, tolerance: float) -> bool: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU: ...
    def ToRotationMatrix(self) -> RotationMatrix_TAutoDiffXdU: ...
    def pitch_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def roll_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def vector(self, *args, **kwargs): ...
    def yaw_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> RollPitchYaw_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RollPitchYaw_TAutoDiffXdU: ...

class RollPitchYaw_𝓣Expression𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self, other: RollPitchYaw_TExpressionU) -> None: ...
    @overload
    def __init__(self, rpy) -> None: ...
    @overload
    def __init__(self, roll: pydrake.symbolic.Expression, pitch: pydrake.symbolic.Expression, yaw: pydrake.symbolic.Expression) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    def CalcAngularVelocityInChildFromRpyDt(self, *args, **kwargs): ...
    def CalcAngularVelocityInParentFromRpyDt(self, *args, **kwargs): ...
    def CalcRotationMatrixDt(self, *args, **kwargs): ...
    def CalcRpyDDtFromAngularAccelInChild(self, *args, **kwargs): ...
    def CalcRpyDDtFromRpyDtAndAngularAccelInParent(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInChild(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInParent(self, *args, **kwargs): ...
    def IsNearlyEqualTo(self, other: RollPitchYaw_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TExpressionU: ...
    def ToRotationMatrix(self) -> RotationMatrix_TExpressionU: ...
    def pitch_angle(self) -> pydrake.symbolic.Expression: ...
    def roll_angle(self) -> pydrake.symbolic.Expression: ...
    def vector(self, *args, **kwargs): ...
    def yaw_angle(self) -> pydrake.symbolic.Expression: ...
    def __copy__(self) -> RollPitchYaw_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RollPitchYaw_TExpressionU: ...

class RotationMatrix:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeFromOneVector: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    MakeXRotation: ClassVar[nanobind.nb_func] = ...
    MakeYRotation: ClassVar[nanobind.nb_func] = ...
    MakeZRotation: ClassVar[nanobind.nb_func] = ...
    ProjectToRotationMatrix: ClassVar[nanobind.nb_func] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    cast𝓣float𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def InvertAndCompose(self, other: RotationMatrix) -> RotationMatrix: ...
    def IsExactlyEqualTo(self, other: RotationMatrix) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RotationMatrix, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, tolerance: float = ...) -> bool: ...
    def IsValid(self) -> bool: ...
    def ToAngleAxis(self) -> pydrake.common.eigen_geometry.AngleAxis: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion: ...
    def ToRollPitchYaw(self) -> RollPitchYaw: ...
    def col(self, *args, **kwargs): ...
    def inverse(self) -> RotationMatrix: ...
    def matrix(self, *args, **kwargs): ...
    def row(self, *args, **kwargs): ...
    def set(self, R) -> None: ...
    def transpose(self) -> RotationMatrix: ...
    def __copy__(self) -> RotationMatrix: ...
    def __deepcopy__(self, arg: dict) -> RotationMatrix: ...

class RotationMatrix_𝓣AutoDiffXd𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeFromOneVector: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    MakeXRotation: ClassVar[nanobind.nb_func] = ...
    MakeYRotation: ClassVar[nanobind.nb_func] = ...
    MakeZRotation: ClassVar[nanobind.nb_func] = ...
    ProjectToRotationMatrix: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def InvertAndCompose(self, other: RotationMatrix_TAutoDiffXdU) -> RotationMatrix_TAutoDiffXdU: ...
    def IsExactlyEqualTo(self, other: RotationMatrix_TAutoDiffXdU) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RotationMatrix_TAutoDiffXdU, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, tolerance: float = ...) -> bool: ...
    def IsValid(self) -> bool: ...
    def ToAngleAxis(self) -> pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU: ...
    def ToRollPitchYaw(self) -> RollPitchYaw_TAutoDiffXdU: ...
    def col(self, *args, **kwargs): ...
    def inverse(self) -> RotationMatrix_TAutoDiffXdU: ...
    def matrix(self, *args, **kwargs): ...
    def row(self, *args, **kwargs): ...
    def set(self, R) -> None: ...
    def transpose(self) -> RotationMatrix_TAutoDiffXdU: ...
    def __copy__(self) -> RotationMatrix_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RotationMatrix_TAutoDiffXdU: ...

class RotationMatrix_𝓣Expression𝓤:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeFromOneVector: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    MakeXRotation: ClassVar[nanobind.nb_func] = ...
    MakeYRotation: ClassVar[nanobind.nb_func] = ...
    MakeZRotation: ClassVar[nanobind.nb_func] = ...
    ProjectToRotationMatrix: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def InvertAndCompose(self, other: RotationMatrix_TExpressionU) -> RotationMatrix_TExpressionU: ...
    def IsExactlyEqualTo(self, other: RotationMatrix_TExpressionU) -> pydrake.symbolic.Formula: ...
    def IsExactlyIdentity(self) -> pydrake.symbolic.Formula: ...
    def IsNearlyEqualTo(self, other: RotationMatrix_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def IsNearlyIdentity(self, tolerance: float = ...) -> pydrake.symbolic.Formula: ...
    def IsValid(self) -> pydrake.symbolic.Formula: ...
    def ToAngleAxis(self) -> pydrake.common.eigen_geometry.AngleAxis_TExpressionU: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TExpressionU: ...
    def ToRollPitchYaw(self) -> RollPitchYaw_TExpressionU: ...
    def col(self, *args, **kwargs): ...
    def inverse(self) -> RotationMatrix_TExpressionU: ...
    def matrix(self, *args, **kwargs): ...
    def row(self, *args, **kwargs): ...
    def set(self, R) -> None: ...
    def transpose(self) -> RotationMatrix_TExpressionU: ...
    def __copy__(self) -> RotationMatrix_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RotationMatrix_TExpressionU: ...

class _TemporaryName_N5drake4math12BsplineBasisINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, order: int, knots: collections.abc.Sequence[pydrake.autodiffutils.AutoDiffXd]) -> None: ...
    @overload
    def __init__(self, order: int, num_basis_functions: int, type: KnotVectorType = ..., initial_parameter_value: pydrake.autodiffutils.AutoDiffXd = ..., final_parameter_value: pydrake.autodiffutils.AutoDiffXd = ...) -> None: ...
    @overload
    def __init__(self, other: BsplineBasis_TAutoDiffXdU) -> None: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_interval: collections.abc.Sequence[pydrake.autodiffutils.AutoDiffXd]) -> list[int]: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> list[int]: ...
    def EvaluateBasisFunctionI(self, i: int, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> pydrake.autodiffutils.AutoDiffXd: ...
    def EvaluateCurve(self, *args, **kwargs): ...
    def EvaluateLinearInControlPoints(self, *args, **kwargs): ...
    def FindContainingInterval(self, parameter_value: pydrake.autodiffutils.AutoDiffXd) -> int: ...
    def degree(self) -> int: ...
    def final_parameter_value(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def initial_parameter_value(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def knots(self) -> list[pydrake.autodiffutils.AutoDiffXd]: ...
    def num_basis_functions(self) -> int: ...
    def order(self) -> int: ...
    def __eq__(self, arg: BsplineBasis_TAutoDiffXdU) -> bool: ...
    def __ne__(self, arg: BsplineBasis_TAutoDiffXdU) -> bool: ...

class _TemporaryName_N5drake4math12BsplineBasisINS_8symbolic10ExpressionEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, order: int, knots: collections.abc.Sequence[pydrake.symbolic.Expression]) -> None: ...
    @overload
    def __init__(self, order: int, num_basis_functions: int, type: KnotVectorType = ..., initial_parameter_value: pydrake.symbolic.Expression = ..., final_parameter_value: pydrake.symbolic.Expression = ...) -> None: ...
    @overload
    def __init__(self, other: BsplineBasis_TExpressionU) -> None: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_interval: collections.abc.Sequence[pydrake.symbolic.Expression]) -> list[int]: ...
    @overload
    def ComputeActiveBasisFunctionIndices(self, parameter_value: pydrake.symbolic.Expression) -> list[int]: ...
    def EvaluateBasisFunctionI(self, i: int, parameter_value: pydrake.symbolic.Expression) -> pydrake.symbolic.Expression: ...
    def EvaluateCurve(self, *args, **kwargs): ...
    def EvaluateLinearInControlPoints(self, *args, **kwargs): ...
    def FindContainingInterval(self, parameter_value: pydrake.symbolic.Expression) -> int: ...
    def degree(self) -> int: ...
    def final_parameter_value(self) -> pydrake.symbolic.Expression: ...
    def initial_parameter_value(self) -> pydrake.symbolic.Expression: ...
    def knots(self) -> list[pydrake.symbolic.Expression]: ...
    def num_basis_functions(self) -> int: ...
    def order(self) -> int: ...
    def __eq__(self, arg: BsplineBasis_TExpressionU) -> pydrake.symbolic.Formula: ...
    def __ne__(self, arg: BsplineBasis_TExpressionU) -> pydrake.symbolic.Formula: ...

class _TemporaryName_N5drake4math12RollPitchYawINS_2ad8AutoDiffEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self, other: RollPitchYaw_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, rpy) -> None: ...
    @overload
    def __init__(self, roll: pydrake.autodiffutils.AutoDiffXd, pitch: pydrake.autodiffutils.AutoDiffXd, yaw: pydrake.autodiffutils.AutoDiffXd) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    def CalcAngularVelocityInChildFromRpyDt(self, *args, **kwargs): ...
    def CalcAngularVelocityInParentFromRpyDt(self, *args, **kwargs): ...
    def CalcRotationMatrixDt(self, *args, **kwargs): ...
    def CalcRpyDDtFromAngularAccelInChild(self, *args, **kwargs): ...
    def CalcRpyDDtFromRpyDtAndAngularAccelInParent(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInChild(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInParent(self, *args, **kwargs): ...
    def IsNearlyEqualTo(self, other: RollPitchYaw_TAutoDiffXdU, tolerance: float) -> bool: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU: ...
    def ToRotationMatrix(self) -> RotationMatrix_TAutoDiffXdU: ...
    def pitch_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def roll_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def vector(self, *args, **kwargs): ...
    def yaw_angle(self) -> pydrake.autodiffutils.AutoDiffXd: ...
    def __copy__(self) -> RollPitchYaw_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RollPitchYaw_TAutoDiffXdU: ...

class _TemporaryName_N5drake4math12RollPitchYawINS_8symbolic10ExpressionEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    @overload
    def __init__(self, other: RollPitchYaw_TExpressionU) -> None: ...
    @overload
    def __init__(self, rpy) -> None: ...
    @overload
    def __init__(self, roll: pydrake.symbolic.Expression, pitch: pydrake.symbolic.Expression, yaw: pydrake.symbolic.Expression) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU) -> None: ...
    @overload
    def __init__(self, matrix) -> None: ...
    def CalcAngularVelocityInChildFromRpyDt(self, *args, **kwargs): ...
    def CalcAngularVelocityInParentFromRpyDt(self, *args, **kwargs): ...
    def CalcRotationMatrixDt(self, *args, **kwargs): ...
    def CalcRpyDDtFromAngularAccelInChild(self, *args, **kwargs): ...
    def CalcRpyDDtFromRpyDtAndAngularAccelInParent(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInChild(self, *args, **kwargs): ...
    def CalcRpyDtFromAngularVelocityInParent(self, *args, **kwargs): ...
    def IsNearlyEqualTo(self, other: RollPitchYaw_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TExpressionU: ...
    def ToRotationMatrix(self) -> RotationMatrix_TExpressionU: ...
    def pitch_angle(self) -> pydrake.symbolic.Expression: ...
    def roll_angle(self) -> pydrake.symbolic.Expression: ...
    def vector(self, *args, **kwargs): ...
    def yaw_angle(self) -> pydrake.symbolic.Expression: ...
    def __copy__(self) -> RollPitchYaw_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RollPitchYaw_TExpressionU: ...

class _TemporaryName_N5drake4math14RigidTransformINS_2ad8AutoDiffEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: RigidTransform_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, rpy: RollPitchYaw_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU, p) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, p) -> None: ...
    @overload
    def __init__(self, pose: pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU) -> None: ...
    @overload
    def __init__(self, pose) -> None: ...
    def GetAsIsometry3(self) -> pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU: ...
    def GetAsMatrix34(self, *args, **kwargs): ...
    def GetAsMatrix4(self, *args, **kwargs): ...
    def GetMaximumAbsoluteDifference(self, other: RigidTransform_TAutoDiffXdU) -> pydrake.autodiffutils.AutoDiffXd: ...
    def GetMaximumAbsoluteTranslationDifference(self, other: RigidTransform_TAutoDiffXdU) -> pydrake.autodiffutils.AutoDiffXd: ...
    def InvertAndCompose(self, other: RigidTransform_TAutoDiffXdU) -> RigidTransform_TAutoDiffXdU: ...
    def IsExactlyEqualTo(self, other: RigidTransform_TAutoDiffXdU) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RigidTransform_TAutoDiffXdU, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, translation_tolerance: float) -> bool: ...
    def SetFromIsometry3(self, pose: pydrake.common.eigen_geometry.Isometry3_TAutoDiffXdU) -> None: ...
    def SetIdentity(self) -> RigidTransform_TAutoDiffXdU: ...
    def inverse(self) -> RigidTransform_TAutoDiffXdU: ...
    def rotation(self) -> RotationMatrix_TAutoDiffXdU: ...
    def set(self, R: RotationMatrix_TAutoDiffXdU, p) -> None: ...
    @overload
    def set_rotation(self, R: RotationMatrix_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, rpy: RollPitchYaw_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU) -> None: ...
    @overload
    def set_rotation(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU) -> None: ...
    def set_translation(self, p) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> RigidTransform_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RigidTransform_TAutoDiffXdU: ...

class _TemporaryName_N5drake4math14RigidTransformINS_8symbolic10ExpressionEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, other: RigidTransform_TExpressionU) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, rpy: RollPitchYaw_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TExpressionU, p) -> None: ...
    @overload
    def __init__(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def __init__(self, p) -> None: ...
    @overload
    def __init__(self, pose: pydrake.common.eigen_geometry.Isometry3_TExpressionU) -> None: ...
    @overload
    def __init__(self, pose) -> None: ...
    def GetAsIsometry3(self) -> pydrake.common.eigen_geometry.Isometry3_TExpressionU: ...
    def GetAsMatrix34(self, *args, **kwargs): ...
    def GetAsMatrix4(self, *args, **kwargs): ...
    def GetMaximumAbsoluteDifference(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Expression: ...
    def GetMaximumAbsoluteTranslationDifference(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Expression: ...
    def InvertAndCompose(self, other: RigidTransform_TExpressionU) -> RigidTransform_TExpressionU: ...
    def IsExactlyEqualTo(self, other: RigidTransform_TExpressionU) -> pydrake.symbolic.Formula: ...
    def IsExactlyIdentity(self) -> pydrake.symbolic.Formula: ...
    def IsNearlyEqualTo(self, other: RigidTransform_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def IsNearlyIdentity(self, translation_tolerance: float) -> pydrake.symbolic.Formula: ...
    def SetFromIsometry3(self, pose: pydrake.common.eigen_geometry.Isometry3_TExpressionU) -> None: ...
    def SetIdentity(self) -> RigidTransform_TExpressionU: ...
    def inverse(self) -> RigidTransform_TExpressionU: ...
    def rotation(self) -> RotationMatrix_TExpressionU: ...
    def set(self, R: RotationMatrix_TExpressionU, p) -> None: ...
    @overload
    def set_rotation(self, R: RotationMatrix_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, rpy: RollPitchYaw_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, quaternion: pydrake.common.eigen_geometry.Quaternion_TExpressionU) -> None: ...
    @overload
    def set_rotation(self, theta_lambda: pydrake.common.eigen_geometry.AngleAxis_TExpressionU) -> None: ...
    def set_translation(self, p) -> None: ...
    def translation(self, *args, **kwargs): ...
    def __copy__(self) -> RigidTransform_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RigidTransform_TExpressionU: ...

class _TemporaryName_N5drake4math14RotationMatrixINS_2ad8AutoDiffEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeFromOneVector: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    MakeXRotation: ClassVar[nanobind.nb_func] = ...
    MakeYRotation: ClassVar[nanobind.nb_func] = ...
    MakeZRotation: ClassVar[nanobind.nb_func] = ...
    ProjectToRotationMatrix: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣AutoDiffXd𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def InvertAndCompose(self, other: RotationMatrix_TAutoDiffXdU) -> RotationMatrix_TAutoDiffXdU: ...
    def IsExactlyEqualTo(self, other: RotationMatrix_TAutoDiffXdU) -> bool: ...
    def IsExactlyIdentity(self) -> bool: ...
    def IsNearlyEqualTo(self, other: RotationMatrix_TAutoDiffXdU, tolerance: float) -> bool: ...
    def IsNearlyIdentity(self, tolerance: float = ...) -> bool: ...
    def IsValid(self) -> bool: ...
    def ToAngleAxis(self) -> pydrake.common.eigen_geometry.AngleAxis_TAutoDiffXdU: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TAutoDiffXdU: ...
    def ToRollPitchYaw(self) -> RollPitchYaw_TAutoDiffXdU: ...
    def col(self, *args, **kwargs): ...
    def inverse(self) -> RotationMatrix_TAutoDiffXdU: ...
    def matrix(self, *args, **kwargs): ...
    def row(self, *args, **kwargs): ...
    def set(self, R) -> None: ...
    def transpose(self) -> RotationMatrix_TAutoDiffXdU: ...
    def __copy__(self) -> RotationMatrix_TAutoDiffXdU: ...
    def __deepcopy__(self, arg: dict) -> RotationMatrix_TAutoDiffXdU: ...

class _TemporaryName_N5drake4math14RotationMatrixINS_8symbolic10ExpressionEEE:
    Identity: ClassVar[nanobind.nb_func] = ...
    MakeFromOneVector: ClassVar[nanobind.nb_func] = ...
    MakeUnchecked: ClassVar[nanobind.nb_func] = ...
    MakeXRotation: ClassVar[nanobind.nb_func] = ...
    MakeYRotation: ClassVar[nanobind.nb_func] = ...
    MakeZRotation: ClassVar[nanobind.nb_func] = ...
    ProjectToRotationMatrix: ClassVar[nanobind.nb_func] = ...
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    cast𝓣Expression𝓤: ClassVar[Callable] = ...
    multiply: ClassVar[Callable] = ...
    __matmul__: ClassVar[Callable] = ...
    cast: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def InvertAndCompose(self, other: RotationMatrix_TExpressionU) -> RotationMatrix_TExpressionU: ...
    def IsExactlyEqualTo(self, other: RotationMatrix_TExpressionU) -> pydrake.symbolic.Formula: ...
    def IsExactlyIdentity(self) -> pydrake.symbolic.Formula: ...
    def IsNearlyEqualTo(self, other: RotationMatrix_TExpressionU, tolerance: float) -> pydrake.symbolic.Formula: ...
    def IsNearlyIdentity(self, tolerance: float = ...) -> pydrake.symbolic.Formula: ...
    def IsValid(self) -> pydrake.symbolic.Formula: ...
    def ToAngleAxis(self) -> pydrake.common.eigen_geometry.AngleAxis_TExpressionU: ...
    def ToQuaternion(self) -> pydrake.common.eigen_geometry.Quaternion_TExpressionU: ...
    def ToRollPitchYaw(self) -> RollPitchYaw_TExpressionU: ...
    def col(self, *args, **kwargs): ...
    def inverse(self) -> RotationMatrix_TExpressionU: ...
    def matrix(self, *args, **kwargs): ...
    def row(self, *args, **kwargs): ...
    def set(self, R) -> None: ...
    def transpose(self) -> RotationMatrix_TExpressionU: ...
    def __copy__(self) -> RotationMatrix_TExpressionU: ...
    def __deepcopy__(self, arg: dict) -> RotationMatrix_TExpressionU: ...
