import pydrake.multibody.plant
import pydrake.multibody.tree
import pydrake.symbolic
from pydrake.common.cpp_template import RationalForwardKinematicsPose_ as RationalForwardKinematicsPose_
from typing import ClassVar

class RationalForwardKinematics:
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant) -> None: ...
    def CalcBodyPoseAsMultilinearPolynomial(self, q_star, body_index: pydrake.multibody.tree.BodyIndex, expressed_body_index: pydrake.multibody.tree.BodyIndex) -> RationalForwardKinematicsPose: ...
    def ComputeQValue(self, *args, **kwargs): ...
    def ComputeSValue(self, *args, **kwargs): ...
    def ConvertMultilinearPolynomialToRationalFunction(self, e: pydrake.symbolic.Polynomial) -> pydrake.symbolic.RationalFunction: ...
    def plant(self) -> pydrake.multibody.plant.MultibodyPlant: ...
    def s(self, *args, **kwargs): ...

class RationalForwardKinematicsPose:
    def __init__(self, *args, **kwargs) -> None: ...
    def position(self, *args, **kwargs): ...
    def rotation(self, *args, **kwargs): ...
    def __copy__(self) -> RationalForwardKinematicsPose: ...
    def __deepcopy__(self, arg: dict) -> RationalForwardKinematicsPose: ...

class RationalForwardKinematicsPose_𝓣RationalFunction𝓤:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def position(self, *args, **kwargs): ...
    def rotation(self, *args, **kwargs): ...
    def __copy__(self) -> RationalForwardKinematicsPose_TRationalFunctionU: ...
    def __deepcopy__(self, arg: dict) -> RationalForwardKinematicsPose_TRationalFunctionU: ...

class _TemporaryName_N5drake9multibody25RationalForwardKinematics4PoseINS_8symbolic16RationalFunctionEEE:
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def position(self, *args, **kwargs): ...
    def rotation(self, *args, **kwargs): ...
    def __copy__(self) -> RationalForwardKinematicsPose_TRationalFunctionU: ...
    def __deepcopy__(self, arg: dict) -> RationalForwardKinematicsPose_TRationalFunctionU: ...
