import pydrake.multibody.contact_solvers
import pydrake.systems.analysis
import pydrake.systems.framework
from pydrake.common.cpp_template import CenicIntegrator_ as CenicIntegrator_
from typing import ClassVar

class CenicIntegrator(pydrake.systems.analysis.IntegratorBase):
    def __init__(self, system: pydrake.systems.framework.System, context: pydrake.systems.framework.Context | None = ...) -> None: ...
    def SetSolverParameters(self, parameters: pydrake.multibody.contact_solvers.IcfSolverParameters) -> None: ...
    def get_solver_parameters(self) -> pydrake.multibody.contact_solvers.IcfSolverParameters: ...

class CenicIntegrator_𝓣AutoDiffXd𝓤(pydrake.systems.analysis.IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...
    def SetSolverParameters(self, parameters: pydrake.multibody.contact_solvers.IcfSolverParameters) -> None: ...
    def get_solver_parameters(self) -> pydrake.multibody.contact_solvers.IcfSolverParameters: ...

class _TemporaryName_N5drake9multibody15CenicIntegratorINS_2ad8AutoDiffEEE(pydrake.systems.analysis.IntegratorBase_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, system: pydrake.systems.framework.System_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU | None = ...) -> None: ...
    def SetSolverParameters(self, parameters: pydrake.multibody.contact_solvers.IcfSolverParameters) -> None: ...
    def get_solver_parameters(self) -> pydrake.multibody.contact_solvers.IcfSolverParameters: ...
