#pragma once

#include "drake/common/drake_copyable.h"

namespace drake {
namespace planning {
namespace experimental {

/** This is a stub class to exercise the build system. */
class Placeholder final {
 public:
  DRAKE_NO_COPY_NO_MOVE_NO_ASSIGN(Placeholder);

  /** This is a stub constructor to exercise the build system. */
  Placeholder();

  ~Placeholder();
};

}  // namespace experimental
}  // namespace planning
}  // namespace drake
